# pylint: disable=W0706
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal
import contextlib
import uuid
import typing

import octobot_commons.symbols as symbol_util
import octobot_commons.constants as commons_constants
import octobot_commons.logging as logging
import octobot_commons.timestamp_util as timestamp_util
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.personal_data.orders.decimal_order_adapter as decimal_order_adapter
import octobot_trading.exchanges.util.exchange_market_status_fixer as exchange_market_status_fixer
import octobot_trading.personal_data.orders.states.fill_order_state as fill_order_state
import octobot_trading.personal_data.orders.order as order_import
import octobot_trading.personal_data.orders.triggers.price_trigger as price_trigger
import octobot_trading.signals as signals
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc


LOGGER_NAME = "order_util"


def is_valid(element, key, zero_valid=False):
    """
    Checks is the element is valid with the market status fixer
    :param element:
    :param key:
    :param zero_valid: if 0 should be considered a valid value
    :return:
    """
    return key in element and exchange_market_status_fixer.is_ms_valid(element[key], zero_valid=zero_valid)


def get_min_max_amounts(symbol_market, default_value=None):
    """
    Returns the min and max quantity, cost and price according to the specified market
    :param symbol_market:
    :param default_value:
    :return:
    """
    min_quantity = max_quantity = min_cost = max_cost = min_price = max_price = default_value
    if Ecmsc.LIMITS.value in symbol_market:
        symbol_market_limits = symbol_market[Ecmsc.LIMITS.value]

        if Ecmsc.LIMITS_AMOUNT.value in symbol_market_limits:
            limit_amount = symbol_market_limits[Ecmsc.LIMITS_AMOUNT.value]
            if is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MIN.value) \
                    or is_valid(limit_amount, Ecmsc.LIMITS_AMOUNT_MAX.value):
                min_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MIN.value, default_value)
                max_quantity = limit_amount.get(Ecmsc.LIMITS_AMOUNT_MAX.value, default_value)

        # case 2: use cost and price
        if Ecmsc.LIMITS_COST.value in symbol_market_limits:
            limit_cost = symbol_market_limits[Ecmsc.LIMITS_COST.value]
            if is_valid(limit_cost, Ecmsc.LIMITS_COST_MIN.value) \
                    or is_valid(limit_cost, Ecmsc.LIMITS_COST_MAX.value):
                min_cost = limit_cost.get(Ecmsc.LIMITS_COST_MIN.value, default_value)
                max_cost = limit_cost.get(Ecmsc.LIMITS_COST_MAX.value, default_value)

        # case 2: use quantity and price
        if Ecmsc.LIMITS_PRICE.value in symbol_market_limits:
            limit_price = symbol_market_limits[Ecmsc.LIMITS_PRICE.value]
            if is_valid(limit_price, Ecmsc.LIMITS_PRICE_MIN.value) \
                    or is_valid(limit_price, Ecmsc.LIMITS_PRICE_MAX.value):
                min_price = limit_price.get(Ecmsc.LIMITS_PRICE_MIN.value, default_value)
                max_price = limit_price.get(Ecmsc.LIMITS_PRICE_MAX.value, default_value)

    return min_quantity, max_quantity, min_cost, max_cost, min_price, max_price


def check_cost(total_order_price, min_cost):
    """
    Checks and adapts the quantity and price of the order to ensure it's exchange compliant:
    - are the quantity and price of the order compliant with the exchange's number of digits requirement
        => otherwise quantity will be truncated accordingly
    - is the quantity valid
    - are the order total price and quantity superior or equal to the exchange's minimum order requirement
        => otherwise order is impossible => returns empty list
    - if total cost data are unavailable:
    - is the price of the currency compliant with the exchange's price interval for this currency
        => otherwise order is impossible => returns empty list
    - are the order total price and quantity inferior or equal to the exchange's maximum order requirement
        => otherwise order is impossible as is => split order into smaller ones and returns the list
    => returns the quantity and price list of possible order(s)
    - if exchange symbol data are not enough
        => try fixing exchange data using ExchangeMarketStatusFixer are start again (once only)
    """
    if total_order_price < min_cost:
        if min_cost is None:
            logging.get_logger(LOGGER_NAME).error("Invalid min_cost from exchange")
        return False
    return True


def get_valid_split_orders(
    quantity: decimal.Decimal, prices: list[decimal.Decimal], symbol_market,
    amount_ratio_per_order: typing.Optional[list[decimal.Decimal]] = None
) -> (list[decimal.Decimal], list[decimal.Decimal]):
    if amount_ratio_per_order:
        if len(amount_ratio_per_order) != len(prices):
            raise ValueError(f"amount_ratio_per_order must have {len(prices)} elements")
        if any(amount_ratio <= constants.ZERO for amount_ratio in amount_ratio_per_order):
            raise ValueError(f"all amount_ratio_per_order must by > {constants.ZERO}")
    if len(prices) < 1:
        return [], []
    if len(prices) < 2:
        return [quantity], [prices[0]]
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = get_min_max_amounts(symbol_market)
    min_quantity = None if min_quantity is None else decimal.Decimal(f"{min_quantity}")
    max_quantity = None if max_quantity is None else decimal.Decimal(f"{max_quantity}")
    min_cost = None if min_cost is None else decimal.Decimal(f"{min_cost}")
    max_cost = None if max_cost is None else decimal.Decimal(f"{max_cost}")
    min_price = None if min_price is None else decimal.Decimal(f"{min_price}")
    max_price = None if max_price is None else decimal.Decimal(f"{max_price}")
    # try to split quantity evenly amount prices
    current_supported_orders_count = decimal.Decimal(str(len(prices)))
    while current_supported_orders_count > constants.ONE:
        if amount_ratio_per_order:
            used_amount_ratio_per_order = amount_ratio_per_order[:int(current_supported_orders_count)]
            total_ratios = sum(used_amount_ratio_per_order)
            candidate_volumes = [
                quantity * ratio_per_order / total_ratios
                for ratio_per_order in used_amount_ratio_per_order
            ]
        else:
            candidate_volumes = [quantity / current_supported_orders_count] * int(current_supported_orders_count)
        candidate_prices = prices[:int(current_supported_orders_count)]
        if _are_orders_too_small(min_quantity, min_cost, min_price, min(candidate_prices), min(candidate_volumes)):
            # reduce orders count to increase quantity
            current_supported_orders_count -= 1
        elif _are_orders_too_large(max_quantity, max_cost, max_price, max(candidate_prices), max(candidate_volumes)):
            raise errors.InvalidArgumentError(f"Order volume ({max(candidate_volumes)}) is too large")
        else:
            return candidate_volumes, candidate_prices
    # default to full quantity and first price
    return [quantity], [prices[0]]


def get_split_orders_count_and_increment(
    lower_price, higher_price, quantity, orders_count, symbol_market, add_increment_to_min_price
) -> (int, decimal.Decimal):
    """
    :param lower_price: smallest order price
    :param higher_price: largest order price
    :param quantity: total quantity to handle
    :param orders_count: ideal orders count
    :param symbol_market: description of the market to trade on
    :param add_increment_to_min_price: when true, uses lower_price + increment to check min values against symbol market
    :return: the exchange compatible orders count and price increment
    """
    min_quantity, max_quantity, min_cost, max_cost, min_price, max_price = get_min_max_amounts(symbol_market)
    min_quantity = None if min_quantity is None else decimal.Decimal(f"{min_quantity}")
    max_quantity = None if max_quantity is None else decimal.Decimal(f"{max_quantity}")
    min_cost = None if min_cost is None else decimal.Decimal(f"{min_cost}")
    max_cost = None if max_cost is None else decimal.Decimal(f"{max_cost}")
    min_price = None if min_price is None else decimal.Decimal(f"{min_price}")
    max_price = None if max_price is None else decimal.Decimal(f"{max_price}")

    limit_check = _ensure_orders_size(
        lower_price, higher_price, quantity, orders_count,
        min_quantity, min_cost, min_price,
        max_quantity, max_cost, max_price,
        symbol_market, add_increment_to_min_price
    )

    while limit_check > 0:
        if limit_check == 1:
            if orders_count > 1:
                orders_count -= 1
            else:
                # not enough funds to create orders
                logging.get_logger(LOGGER_NAME).warning(f"Not enough funds to create order.")
                return 0, constants.ZERO
        elif limit_check == 2:
            if orders_count < 40:
                orders_count += 1
            else:
                # too many orders to create, must be a problem
                logging.get_logger(LOGGER_NAME).error("Too many orders to create.")
                return 0, constants.ZERO
        limit_check = _ensure_orders_size(
            lower_price, higher_price, quantity, orders_count,
            min_quantity, min_cost, min_price,
            max_quantity, max_cost, max_price,
            symbol_market, add_increment_to_min_price
        )
    return orders_count, (higher_price - lower_price) / orders_count


def _ensure_orders_size(lower_price, higher_price, quantity, orders_count,
                        min_quantity, min_cost, min_price,
                        max_quantity, max_cost, max_price,
                        symbol_market, add_increment_to_min_price):
    increment = (higher_price - lower_price) / orders_count
    first_price = (lower_price + increment) if add_increment_to_min_price else lower_price
    last_price = lower_price + (increment * orders_count)
    order_vol = decimal_order_adapter.decimal_adapt_quantity(symbol_market, quantity / orders_count)

    if _are_orders_too_small(min_quantity, min_cost, min_price, first_price, order_vol):
        return 1
    elif _are_orders_too_large(max_quantity, max_cost, max_price, last_price, order_vol):
        return 2
    return 0


def _are_orders_too_small(min_quantity, min_cost, min_price, price, volume):
    return (min_price and price < min_price) or \
           (min_quantity and volume < min_quantity) or \
           (min_cost and price * volume < min_cost)


def _are_orders_too_large(max_quantity, max_cost, max_price, price, volume):
    return (max_price and price > max_price) or \
           (max_quantity and volume > max_quantity) or \
           (max_cost and price * volume > max_cost)


async def get_up_to_date_price(exchange_manager, symbol: str, timeout: int = None, base_error: str = None):
    exchange_time = exchange_manager.exchange.get_exchange_current_time()
    base_error = base_error or f"Can't get the necessary {exchange_manager.exchange_name} " \
                               f"price data to create a new {symbol} order on the " \
                               f"{timestamp_util.convert_timestamp_to_datetime(exchange_time)} " \
                               f"(timestamp: {exchange_time}):"
    try:
        mark_price = await exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol) \
            .prices_manager.get_mark_price(timeout=timeout)
    except asyncio.TimeoutError:
        raise asyncio.TimeoutError(f"{base_error} mark price is not available")
    except errors.UnreachableExchange as e:
        raise errors.UnreachableExchange(f"{base_error} exchange is unreachable") from e
    return decimal.Decimal(str(mark_price))


def get_potentially_outdated_price(exchange_manager, symbol: str) -> (decimal.Decimal, bool):
    symbol_data = exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        symbol, allow_creation=False
    )
    try:
        return symbol_data.prices_manager.get_mark_price_no_wait(), True
    except ValueError:
        # outdated price, return it anyway
        return symbol_data.prices_manager.mark_price, False


async def get_pre_order_data(exchange_manager, symbol: str, timeout: int = None,
                             portfolio_type=commons_constants.PORTFOLIO_AVAILABLE,
                             target_price=None):
    price = target_price or await get_up_to_date_price(exchange_manager, symbol, timeout=timeout)
    symbol_market = exchange_manager.exchange.get_market_status(symbol, with_fixer=False)
    currency_available, market_available, market_quantity = get_portfolio_amounts(
        exchange_manager, symbol, price, portfolio_type=portfolio_type
    )
    return currency_available, market_available, market_quantity, price, symbol_market


def get_portfolio_amounts(exchange_manager, symbol, price, portfolio_type=commons_constants.PORTFOLIO_AVAILABLE):
    currency, market = symbol_util.parse_symbol(symbol).base_and_quote()
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio
    currency_available = portfolio.get_currency_portfolio(currency).available \
        if portfolio_type == commons_constants.PORTFOLIO_AVAILABLE else portfolio.get_currency_portfolio(currency).total
    market_available = portfolio.get_currency_portfolio(market).available \
        if portfolio_type == commons_constants.PORTFOLIO_AVAILABLE else portfolio.get_currency_portfolio(market).total

    if exchange_manager.is_future:
        pair_future_contract = exchange_manager.exchange.get_pair_future_contract(symbol)
        if pair_future_contract.is_inverse_contract():
            currency_available *= pair_future_contract.current_leverage
            market_quantity = market_available = currency_available
        else:
            market_available *= pair_future_contract.current_leverage / price
            market_quantity = currency_available = market_available
    elif exchange_manager.is_margin:
        market_quantity = constants.ZERO  # TODO
    else:
        market_quantity = market_available / price if price else constants.ZERO
    return currency_available, market_available, market_quantity


def get_futures_max_order_size(exchange_manager, symbol, side, current_price, reduce_only,
                               current_symbol_holding, market_quantity):
    # use position margin when trading futures and reducing the position
    current_position = exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
        symbol,
        enums.PositionSide.BOTH
    )
    if reduce_only and current_position.is_idle():
        # can't reduce an empty position
        return constants.ZERO, False
    # ensure max position order size is taken into account
    new_position_side = current_position.side
    if new_position_side is enums.PositionSide.UNKNOWN:
        new_position_side = enums.PositionSide.LONG if side is enums.TradeOrderSide.BUY \
            else enums.PositionSide.SHORT

    contract_current_symbol_holding = current_symbol_holding
    contract_market_quantity = market_quantity
    added_quantity_from_reverse = constants.ZERO
    might_reverse_position = False
    if current_position.symbol_contract.is_inverse_contract():
        # use USD (in BTC/USD) for order sizes on inverse, convert BTC values into USD ones
        contract_current_symbol_holding = current_symbol_holding * current_price
        contract_market_quantity = market_quantity * current_price
    if side is enums.TradeOrderSide.SELL and current_position.is_long():
        # can also sell the position size in long
        contract_current_symbol_holding = current_position.size
        might_reverse_position = not reduce_only
        added_quantity_from_reverse = current_position.initial_margin
    elif side is enums.TradeOrderSide.BUY and current_position.is_short():
        # can also buy the position size in short
        contract_market_quantity = abs(current_position.size)
        might_reverse_position = not reduce_only
        added_quantity_from_reverse = current_position.initial_margin
    max_reducing_position_order_size = (
        contract_market_quantity if side is enums.TradeOrderSide.BUY else contract_current_symbol_holding
    )
    if (
        (new_position_side is enums.PositionSide.LONG and side is enums.TradeOrderSide.BUY)
        or (new_position_side is enums.PositionSide.SHORT and  side is enums.TradeOrderSide.SELL)
        or might_reverse_position
    ):
        quantity = market_quantity if current_position.symbol_contract.is_inverse_contract() \
            else market_quantity * current_price
        # include added_quantity_from_reverse in case funds get freed from reversing a position
        usable_quantity = quantity + added_quantity_from_reverse
        unleveraged_quantity = usable_quantity / current_position.symbol_contract.current_leverage
        max_position_increased_order_quantity = get_max_order_quantity_for_price(
            current_position, unleveraged_quantity, current_price, new_position_side, symbol
        )
        # apply MAX_INCREASED_POSITION_QUANTITY_MULTIPLIER in case the total order cost computation
        # is not (yet) accurate on this exchange (default is 1, meaning the calculation is accurate)
        if exchange_manager.exchange.MAX_INCREASED_POSITION_QUANTITY_MULTIPLIER != constants.ONE \
           and not exchange_manager.is_backtesting:
            max_position_increased_order_quantity *= \
                exchange_manager.exchange.MAX_INCREASED_POSITION_QUANTITY_MULTIPLIER
        # increasing position: always use the same currency
        return max_position_increased_order_quantity + (
            # include reducing position amount to process reverse
            max_reducing_position_order_size if might_reverse_position else constants.ZERO
        ), True
    return max_reducing_position_order_size, False


def get_max_order_quantity_for_price(position, available_quantity, price, side, symbol):
    """
    Returns the maximum order quantity in market or currency for given total usable funds, price and side.
    This amount is not the total usable funds as it also requires to keep the position's open order fees
    as well as the potential position liquidation fees in portfolio. Those fees are computed by
    get_two_way_taker_fee_for_quantity_and_price
    :param position: the position to compute quantity for
    :param available_quantity: the maximum amount of currency/market to allocate to the position (without leverage)
    :param price: the target entry price of the position
    :param side: the side of the position
    :param side: the symbol of the position
    :return: the computed leveraged maximum entry quantity
    """
    # use position.symbol_contract.current_leverage as quantity to simulate a 1 unit quantity (x leverage)
    two_way_fees = position.get_two_way_taker_fee_for_quantity_and_price(position.symbol_contract.current_leverage,
                                                                         price, side, symbol)
    if position.symbol_contract.is_inverse_contract():
        # Returns the maximum order quantity in market.
        return position.symbol_contract.current_leverage * available_quantity / \
            (two_way_fees + constants.ONE / price)
    # Returns the maximum order quantity in currency.
    return position.symbol_contract.current_leverage * available_quantity / \
        (two_way_fees + price)


def get_locked_funds(order):
    forecasted_fees = order.get_computed_fee(use_origin_quantity_and_price=not order.is_filled())
    if order.side == enums.TradeOrderSide.BUY:
        # locking quote to buy
        quote_fees = get_fees_for_currency(forecasted_fees, order.market)
        return order.origin_quantity * order.origin_price + quote_fees
    else:
        # locking base to sell
        base_fees = get_fees_for_currency(forecasted_fees, order.currency)
        return order.origin_quantity + base_fees


def total_fees_from_order_dict(order_dict, currency):
    return get_fees_for_currency(order_dict[enums.ExchangeConstantsOrderColumns.FEE.value], currency)


def get_fees_for_currency(fee, currency):
    if fee and fee[enums.FeePropertyColumns.CURRENCY.value] == currency:
        return decimal.Decimal(str(fee[enums.FeePropertyColumns.COST.value]))
    return constants.ZERO


def get_order_locked_amount(order: order_import.Order) -> decimal.Decimal:
    # take fees into account when in locked asset
    # ( a BTC/USDT order with USDT fees need to lock USDT fees to be able to pay them)
    forecasted_fees = order.get_computed_fee(use_origin_quantity_and_price=not order.is_filled())
    base, quote = symbol_util.parse_symbol(order.symbol).base_and_quote()
    # when buy order
    if order.side == enums.TradeOrderSide.BUY:
        return order.origin_quantity * order.origin_price + get_fees_for_currency(forecasted_fees, quote)
    # when sell order
    return order.origin_quantity + get_fees_for_currency(forecasted_fees, base)


def get_orders_locked_amounts_by_asset(open_orders: list[order_import.Order]) -> dict[str, decimal.Decimal]:
    if not open_orders:
        return {}
    locked_funds_by_asset = {}
    for order in open_orders:
        base, quote = symbol_util.parse_symbol(order.symbol).base_and_quote()
        # use get_order_locked_amount just like trader simulator to ensure locked funds integrity
        if order.side == enums.TradeOrderSide.BUY:
            # buy orders only lock fees in quote
            locked_quote = get_order_locked_amount(order)
            if quote not in locked_funds_by_asset:
                locked_funds_by_asset[quote] = locked_quote
            else:
                locked_funds_by_asset[quote] += locked_quote
        else:
            # sell orders only lock fees in base
            locked_base = get_order_locked_amount(order)
            if base not in locked_funds_by_asset:
                locked_funds_by_asset[base] = locked_base
            else:
                locked_funds_by_asset[base] += locked_base
    return locked_funds_by_asset


def parse_raw_fees(raw_fees):
    fees = raw_fees
    if fees:
        # parsed fees should be from exchange by default
        fees[enums.FeePropertyColumns.IS_FROM_EXCHANGE.value] = \
            fees.get(enums.FeePropertyColumns.IS_FROM_EXCHANGE.value, True)
        if enums.FeePropertyColumns.COST.value in fees:
            try:
                raw_fees[enums.FeePropertyColumns.COST.value] = \
                    decimal.Decimal(str(raw_fees[enums.FeePropertyColumns.COST.value]))
            except decimal.InvalidOperation:
                # Ensure fee cost can be used in computations. The original value is kept
                # under the EXCHANGE_ORIGINAL_COST key if relevant
                raw_fees[enums.FeePropertyColumns.COST.value] = constants.ZERO
    return fees


def parse_order_status(raw_order):
    try:
        return enums.OrderStatus(raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value])
    except KeyError:
        return enums.OrderStatus.UNKNOWN
    except ValueError:
        if raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value] == "cancelled":
            # few exchanges use "cancelled" which is not in enums.ExchangeConstantsOrderColumns.STATUS
            raw_order[enums.ExchangeConstantsOrderColumns.STATUS.value] = enums.OrderStatus.CANCELED.value
            return enums.OrderStatus.CANCELED
        raise


def parse_is_cancelled(raw_order):
    return parse_order_status(raw_order) in {enums.OrderStatus.CANCELED, enums.OrderStatus.CLOSED}


def parse_is_pending_cancel(raw_order):
    return parse_order_status(raw_order) is enums.OrderStatus.PENDING_CANCEL


def parse_is_open(raw_order):
    return parse_order_status(raw_order) is enums.OrderStatus.OPEN


def get_pnl_transaction_source_from_order(order):
    if order.order_type in [enums.TraderOrderType.SELL_MARKET, enums.TraderOrderType.BUY_MARKET,
                            enums.TraderOrderType.TAKE_PROFIT]:
        return enums.PNLTransactionSource.MARKET_ORDER
    if order.order_type in [enums.TraderOrderType.SELL_LIMIT, enums.TraderOrderType.BUY_LIMIT,
                            enums.TraderOrderType.TAKE_PROFIT_LIMIT]:
        return enums.PNLTransactionSource.LIMIT_ORDER
    if is_stop_order(order.order_type):
        return enums.PNLTransactionSource.STOP_ORDER
    return enums.PNLTransactionSource.UNKNOWN


def is_stop_order(order_type: enums.TraderOrderType):
    return order_type in [
        enums.TraderOrderType.STOP_LOSS, enums.TraderOrderType.STOP_LOSS_LIMIT,
        enums.TraderOrderType.TRAILING_STOP, enums.TraderOrderType.TRAILING_STOP_LIMIT,
    ]


def is_stop_trade_order_type(order_type: enums.TradeOrderType):
    return order_type in [
        enums.TradeOrderType.STOP_LOSS, enums.TradeOrderType.STOP_LOSS_LIMIT,
        enums.TradeOrderType.TRAILING_STOP, enums.TradeOrderType.TRAILING_STOP_LIMIT,
    ]


def is_take_profit_order(order_type: enums.TraderOrderType):
    return order_type in [
        enums.TraderOrderType.TAKE_PROFIT, enums.TraderOrderType.TAKE_PROFIT_LIMIT,
    ]


def get_trade_order_type(order_type: enums.TraderOrderType) -> enums.TradeOrderType:
    if order_type in (enums.TraderOrderType.BUY_MARKET, enums.TraderOrderType.SELL_MARKET):
        return enums.TradeOrderType.MARKET
    if order_type in (enums.TraderOrderType.BUY_LIMIT, enums.TraderOrderType.SELL_LIMIT):
        return enums.TradeOrderType.LIMIT
    if order_type is enums.TraderOrderType.STOP_LOSS:
        return enums.TradeOrderType.STOP_LOSS
    if order_type is enums.TraderOrderType.TRAILING_STOP:
        return enums.TradeOrderType.TRAILING_STOP
    if order_type is enums.TraderOrderType.STOP_LOSS_LIMIT:
        return enums.TradeOrderType.STOP_LOSS_LIMIT
    if order_type is enums.TraderOrderType.TRAILING_STOP_LIMIT:
        return enums.TradeOrderType.TRAILING_STOP_LIMIT
    if order_type is enums.TraderOrderType.TAKE_PROFIT:
        return enums.TradeOrderType.TAKE_PROFIT
    if order_type is enums.TraderOrderType.TAKE_PROFIT_LIMIT:
        return enums.TradeOrderType.TAKE_PROFIT_LIMIT
    raise ValueError(order_type)


def create_order_price_trigger(
    order: order_import.Order, active_trigger_price: decimal.Decimal, active_trigger_above: bool
) -> price_trigger.PriceTrigger:
    if active_trigger_price is None or active_trigger_above is None:
        raise ValueError("active_trigger_price and active_trigger_above must be specified")
    return price_trigger.PriceTrigger(
        order.on_active_trigger, (None, None), active_trigger_price, active_trigger_above
    )


async def create_as_active_order_using_strategy_if_any(
    order, strategy_timeout: typing.Optional[float], wait_for_fill_callback: typing.Optional[typing.Callable]
):
    if active_swap_strategy := order.order_group.active_order_swap_strategy if order.order_group else None:
        # use strategy to process swap
        await active_swap_strategy.execute(order, wait_for_fill_callback, strategy_timeout)
    else:
        # no strategy: just create this order as active
        await create_as_active_order_on_exchange(order, False)


async def create_as_active_order_on_exchange(order_to_create, emit_trading_signals: bool):
    async with signals.remote_signal_publisher(
        order_to_create.trader.exchange_manager, order_to_create.symbol, emit_trading_signals
    ):
        active_order = await signals.update_order_as_active(
            order_to_create.trader.exchange_manager,
            signals.should_emit_trading_signal(order_to_create.trader.exchange_manager),
            order_to_create,
            params=order_to_create.exchange_creation_params
        )
        if active_order is None:
            logging.get_logger(order_to_create.get_logger_name()).error(
                f"Failed to create active order for {order_to_create}"
            )
        return active_order


async def update_order_as_inactive_on_exchange(order_to_cancel, emit_trading_signals: bool):
    async with signals.remote_signal_publisher(
        order_to_cancel.trader.exchange_manager, order_to_cancel.symbol, emit_trading_signals
    ):
        if not await signals.update_order_as_inactive(
            order_to_cancel.trader.exchange_manager,
            signals.should_emit_trading_signal(order_to_cancel.trader.exchange_manager),
            order_to_cancel,
        ):
            logging.get_logger(order_to_cancel.get_logger_name()).error(f"Failed to cancel active order on exchange: {order_to_cancel}")
            return False
    return True


async def create_as_chained_order(order):
    order.is_waiting_for_chained_trigger = False
    if not order.trader.simulate and order.has_been_bundled:
        # exchange should have created it already, it is either already fetched or
        # will automatically be fetched at the next update
        # warning: not handling instantly filled bundled orders as there is no easy way to do this
        # TODO: figure out instantly filled bundled orders
        if not await _apply_pending_order_on_existing_orders(order):
            # register it as pending creation order for it to be found and update when fetched
            order.exchange_manager.exchange_personal_data.orders_manager.register_pending_creation_order(order)
    else:
        # set created now to consider creation failures as created as well (the caller can always retry later on)
        order.status = enums.OrderStatus.OPEN
        # set uninitialized to allow second initialization from create_order
        order.is_initialized = False
        order.creation_time = order.exchange_manager.exchange.get_exchange_current_time()
        try:
            await order.trader.create_order(
                order,
                loaded=False,
                params=order.exchange_creation_params,
                raise_all_creation_error=True,
                **order.trader_creation_kwargs
            )
        except (errors.ExchangeClosedPositionError, errors.ExchangeOrderInstantTriggerError):
            # Order can be created and might be outdated forward error for the caller to fix it if possible
            raise
        except Exception as err:
            # log warning to be sure to keep track of the failed order details
            logging.get_logger(LOGGER_NAME).warning(
                f"Failed to create chained order {order.to_dict()}: {err} ({err.__class__.__name__})"
            )
            # propagate
            raise


def is_associated_pending_order(pending_order, created_order):
    return created_order.exchange_order_id == pending_order.exchange_order_id or (
        created_order.symbol == pending_order.symbol and
        created_order.origin_quantity == pending_order.origin_quantity and
        created_order.origin_price == pending_order.origin_price and
        created_order.__class__ is pending_order.__class__ and
        created_order.trader is pending_order.trader
    )


async def apply_pending_order_from_created_order(pending_order, created_order, to_be_initialized):
    await pending_order.update_from_order(created_order)
    if to_be_initialized:
        pending_order.is_initialized = False
    logging.get_logger(LOGGER_NAME).debug(f"Updated pending order: {pending_order} using {created_order}")


async def _apply_pending_order_on_existing_orders(pending_order):
    for created_order in pending_order.exchange_manager.exchange_personal_data.orders_manager.get_open_orders(
        symbol=pending_order.symbol
    ):
        if is_associated_pending_order(pending_order, created_order) and created_order.order_group is None:
            await apply_pending_order_from_created_order(pending_order, created_order, False)
            pending_order.exchange_manager.exchange_personal_data.orders_manager.replace_order(
                created_order.order_id, pending_order
            )
            created_order.clear()
            return True
    return False


@contextlib.asynccontextmanager
async def ensure_orders_relevancy(order=None, position=None, enable_associated_orders_creation=True):
    exchange_manager = order.exchange_manager if position is None else position.exchange_manager
    # part used in futures trading only
    if exchange_manager.exchange_personal_data.positions_manager.positions:
        position = position or exchange_manager.exchange_personal_data.positions_manager.get_order_position(order)
        pre_update_position_side = position.side
        is_pre_update_position_idle = position.is_idle()
        yield
        if not is_pre_update_position_idle and \
           (position.side != pre_update_position_side or position.is_idle()):
            # when position side is changing (from a non-idle position) or is going back to idle,
            # then associated reduce only orders must be closed
            await _cancel_reduce_only_orders_on_position_reset(
                exchange_manager, position.symbol, enable_associated_orders_creation
            )
    else:
        # as a context manager, yield is mandatory
        yield


async def _cancel_reduce_only_orders_on_position_reset(exchange_manager, symbol, enable_associated_orders_creation):
    for order in list(exchange_manager.exchange_personal_data.orders_manager.get_open_orders(symbol)):
        # reduce only order are automatically cancelled on exchanges, only cancel simulated orders
        if (exchange_manager.is_trader_simulated or order.is_self_managed()) \
                and order.is_open() and order.reduce_only:
            try:
                await order.trader.cancel_order(order)
                if order.order_group and enable_associated_orders_creation:
                    await order.order_group.on_cancel(order)
            except (    # pylint: disable=try-except-raise
                errors.OrderCancelError, errors.UnexpectedExchangeSideOrderStateError
            ):
                # should never happen as those should be simulated orders
                raise


def get_order_quantity_currency(exchange_manager, symbol):
    try:
        parsed_symbol = symbol_util.parse_symbol(symbol)
        base, quote = parsed_symbol.base_and_quote()
    except ValueError:
        # symbol that can't be split
        return None
    if exchange_manager.is_future:
        return quote if parsed_symbol.is_inverse() else base
    # always base in spot
    return base


async def get_order_size_portfolio_percent(exchange_manager, order_amount, side, symbol):
    current_symbol_holding, current_market_holding, market_quantity, _, _ = \
        await get_pre_order_data(exchange_manager,
                                 symbol=symbol,
                                 timeout=constants.ORDER_DATA_FETCHING_TIMEOUT,
                                 portfolio_type=commons_constants.PORTFOLIO_TOTAL)
    if exchange_manager.is_future:
        # TODO check inverse
        if market_quantity == constants.ZERO:
            return constants.ZERO
        return min(order_amount / market_quantity, constants.ONE) * constants.ONE_HUNDRED
    if side is enums.TradeOrderSide.SELL:
        if current_symbol_holding == constants.ZERO:
            return constants.ZERO
        return min(order_amount / current_symbol_holding, constants.ONE) * constants.ONE_HUNDRED
    if side is enums.TradeOrderSide.BUY:
        if current_market_holding == constants.ZERO:
            return constants.ZERO
        return min(order_amount / market_quantity, constants.ONE) * constants.ONE_HUNDRED
    raise errors.InvalidArgumentError(f"Unhandled side: {side}")


def generate_order_id():
    return str(uuid.uuid4())


def _get_possible_filled_price(
    exchange_manager, symbol: str, side: enums.TradeOrderSide, ideal_price: decimal.Decimal
) -> decimal.Decimal:
    try:
        min_price, max_price = exchange_manager.exchange_symbols_data.get_exchange_symbol_data(symbol).\
            price_events_manager.get_min_and_max_prices()
        if side is enums.TradeOrderSide.SELL:
            if ideal_price < min_price:
                return min_price
        if side is enums.TradeOrderSide.BUY:
            if ideal_price > max_price:
                return max_price
    except IndexError:
        # no available price data
        pass
    return ideal_price


def get_valid_filled_price(order, ideal_price: decimal.Decimal):
    # ensure filled price based on ideal_price makes sense according to the current candle
    # instead of blindly using ideal_price as this could
    # potentially buy higher than the candle high or sell lower than the candle low,
    # which can't happen on real conditions

    if order.exchange_manager.is_backtesting:
        # in backtesting: ensure ideal_price is with in current candle price.
        return _get_possible_filled_price(order.exchange_manager, order.symbol, order.side, ideal_price)
    # not in backtesting: price desync can't happen
    return ideal_price


async def adapt_chained_order_before_creation(base_order, chained_order):
    can_be_created = True
    if chained_order.update_with_triggering_order_fees:
        can_be_created = chained_order.update_quantity_with_order_fees(base_order)
    # ensure price is not outdated
    await chained_order.update_price_if_outdated()
    return can_be_created


async def wait_for_order_fill(order, timeout, wait_for_portfolio_update):
    if order.is_open():
        if order.state is None:
            logging.get_logger(LOGGER_NAME).error(
                f"None state on created order, impossible to wait for fill, order: {order}"
            )
        else:
            try:
                await order.state.wait_for_next_state(timeout)
            except asyncio.TimeoutError:
                logging.get_logger(LOGGER_NAME).error(
                    f"Timeout while waiting for {order.order_type.value} open order fill, order {order}"
                )
            if wait_for_portfolio_update and isinstance(order.state, fill_order_state.FillOrderState):
                # portfolio is updated in FillOrderState: wait for this state to complete
                try:
                    await order.state.wait_for_next_state(timeout)
                except asyncio.TimeoutError:
                    logging.get_logger(LOGGER_NAME).error(
                        f"Timeout while waiting for {order.order_type.value} filled order state to "
                        f"complete, order {order}"
                    )
    if order.is_open():
        logging.get_logger(LOGGER_NAME).error(f"Unexpected: order is still open, order {order}")
    else:
        logging.get_logger(LOGGER_NAME).info(f"Successfully filled order: {order}.")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import dataclasses
import decimal
import typing

import octobot_commons.logging as commons_logging
import octobot_commons.constants as commons_constants
import octobot_trading.personal_data as personal_data
import octobot_trading.constants as constants


@dataclasses.dataclass
class SubPortfolioData:
    bot_id: typing.Optional[str]
    portfolio_id: typing.Optional[str]
    priority_key: float
    content: dict[str, dict[str, decimal.Decimal]]  # current content of the sub-portfolio
    unit: typing.Optional[str]
    # deltas to be applied on top of the current content of the sub-portfolio from get_content_after_deltas()
    funds_deltas: dict[str, dict[str, decimal.Decimal]] = dataclasses.field(default_factory=dict)
    # funds that are missing from this portfolio. Populated after portfolio has been resolved
    missing_funds: dict[str, decimal.Decimal] = dataclasses.field(default_factory=dict)

    # funds that are locked (maybe in orders) from this portfolio
    locked_funds_by_asset: dict[str, decimal.Decimal] = dataclasses.field(default_factory=dict)


    def get_content_after_deltas(self) -> dict[str, dict[str, decimal.Decimal]]:
        """
        Computes the updated portfolio from delta available and total
        """
        updated_content = copy.deepcopy(self.content)
        for asset, delta_values in self.funds_deltas.items():
            if asset in updated_content:
                updated_content[asset][commons_constants.PORTFOLIO_TOTAL] += (
                    delta_values[commons_constants.PORTFOLIO_TOTAL]
                )
                updated_content[asset][commons_constants.PORTFOLIO_AVAILABLE] += (
                    delta_values[commons_constants.PORTFOLIO_AVAILABLE]
                )
            else:
                updated_content[asset] = {
                    commons_constants.PORTFOLIO_TOTAL: delta_values[commons_constants.PORTFOLIO_TOTAL],
                    commons_constants.PORTFOLIO_AVAILABLE: delta_values[commons_constants.PORTFOLIO_AVAILABLE],
                }
            for key in (commons_constants.PORTFOLIO_TOTAL, commons_constants.PORTFOLIO_AVAILABLE):
                if updated_content[asset][key] < constants.ZERO:
                    commons_logging.get_logger(__name__).warning(
                        f"{asset} portfolio {key} value: {updated_content[asset][key]} is < 0. Replacing with zero."
                    )
                    updated_content[asset][key] = constants.ZERO
        return updated_content

    def get_content_from_total_deltas_and_locked_funds(self) -> dict[str, dict[str, decimal.Decimal]]:
        """
        Only uses PORTFOLIO_TOTAL deltas and computes available values from locked_funds_by_asset
        """
        updated_content = copy.deepcopy(self.content)
        for asset, delta_values in self.funds_deltas.items():
            if asset in updated_content:
                updated_content[asset][commons_constants.PORTFOLIO_TOTAL] += (
                    delta_values[commons_constants.PORTFOLIO_TOTAL]
                )
            else:
                updated_content[asset] = {
                    commons_constants.PORTFOLIO_TOTAL: delta_values[commons_constants.PORTFOLIO_TOTAL]
                }
        for asset, values in updated_content.items():
            locked_funds = self.locked_funds_by_asset.get(asset, constants.ZERO)
            if locked_funds > values[commons_constants.PORTFOLIO_TOTAL]:
                commons_logging.get_logger(__name__).error(
                    f"Unexpected: negative {asset} available value after applying {locked_funds} locked funds to {values}"
                )
            values[commons_constants.PORTFOLIO_AVAILABLE] = values[commons_constants.PORTFOLIO_TOTAL] - locked_funds
        return updated_content

    def is_similar_to(self, other) -> bool:
        return (
            other
            and personal_data.filter_empty_values(self.content) == personal_data.filter_empty_values(other.content)
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order as order_class


class MarketOrder(order_class.Order):
    SUPPORTS_GROUPING = False    # False when orders of this type can't be grouped

    async def update_order_status(self, force_refresh=False):
        if not self.is_active:
            octobot_commons.logging.get_logger(self.logger_name).error(
                f"UNEXPECTED: a {self.get_name()} can't be inactive"
            )
            return
        if self.trader.simulate:
            # TODO: ensure no issue un not running it in task anymore
            await self.on_fill(force_fill=True)
            # asyncio.create_task(self.on_fill(force_fill=True))
            # # In trading simulation wait for the next asyncio loop iteration to ensure this order status
            # # is updated before leaving this method
            # await asyncio_tools.wait_asyncio_next_cycle()

    def on_fill_actions(self):
        self.taker_or_maker = enums.ExchangeConstantsMarketPropertyColumns.TAKER.value
        self.origin_price = self.created_last_price
        self.update_order_filled_values(self.created_last_price)
        order_class.Order.on_fill_actions(self)

    def _should_instant_fill(self):
        return True

    def can_be_edited(self):
        # instantly filled orders can't be edited
        return False

    def use_current_price_as_origin_price(self):
        # Override to return True when the current order price can't be set by the user (ex: market orders)
        return True




#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import octobot_trading.personal_data.orders.order_group as order_group
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.active_order_swap_strategies as active_order_swap_strategies
import octobot_trading.errors as errors
import octobot_trading.signals as signals


class OneCancelsTheOtherOrderGroup(order_group.OrderGroup):
    """
    OneCancelsTheOtherOrderGroup is linking orders together in the way that if any order of the group is filled
    order cancelled, all the orders are cancelled.
    Mostly used to pair stop and limit orders together in a stop-loss / take-profit setup
    """

    async def on_fill(self, filled_order, ignored_orders=None):
        """
        Called when an order referencing this group is filled
        This is called right before updating portfolio for this filled order and the
        order fill publication
        :param filled_order: the filled order
        :param ignored_orders: orders that should be ignored
        """
        if not self.enabled:
            return
        await self._cancel_orders(filled_order, "filled", filled_order)

    async def on_cancel(self, cancelled_order, ignored_orders=None):
        """
        Called when an order referencing this group is cancelled
        This is called before updating portfolio for this cancelled order and the
        order cancel publication
        :param cancelled_order: the cancelled order
        :param ignored_orders: orders that should be ignored
        """
        if not self.enabled:
            return
        if ignored_orders and len(ignored_orders) > 1:
            raise errors.OrderGroupTriggerArgumentError(f"ignored_orders supports at most 1 argument "
                                                        f"for {self.__class__.__name__}")
        ignored_order = ignored_orders[0] if ignored_orders else None
        await self._cancel_orders(cancelled_order, "cancelled", ignored_order)

    async def adapt_before_order_becoming_active(self, order_to_become_active) -> (list, typing.Callable[[list], None]):
        """
        Called before an order referencing this group is becoming active
        """
        # convert the other order of this group into an inactive order => cancel it on exchange
        deactivated_orders = []
        for order in self.get_group_open_orders():
            if order is not order_to_become_active:
                self.logger.info(
                    f"Cancelling order [{order}] from order group as paired order "
                    f"is becoming active ({order_to_become_active})"
                )
                if await order_util.update_order_as_inactive_on_exchange(order, False):
                    deactivated_orders.append(order)
        return deactivated_orders, self._reverse_active_swap

    async def _reverse_active_swap(self, former_order_to_become_active, to_activate_orders):
        for order in to_activate_orders:
            await self.adapt_before_order_becoming_active(order)
            await order_util.create_as_active_order_on_exchange(order, False)

    def _default_active_order_swap_strategy(self, timeout: float) -> active_order_swap_strategies.ActiveOrderSwapStrategy:
        """
        Called when an order of this group is becoming active
        """
        return active_order_swap_strategies.StopFirstActiveOrderSwapStrategy(timeout)

    async def _cancel_orders(self, triggering_order, trigger, ignored_order):
        for order in self.get_group_open_orders():
            if order is not triggering_order and order.is_open():
                try:
                    self.logger.info(f"Cancelling order [{order}] from order group as {triggering_order} is {trigger}")
                    async with signals.remote_signal_publisher(order.trader.exchange_manager, order.symbol, True):
                        await signals.cancel_order(
                            order.trader.exchange_manager,
                            signals.should_emit_trading_signal(order.trader.exchange_manager),
                            order,
                            ignored_order=ignored_order
                        )
                except (errors.OrderCancelError, errors.UnexpectedExchangeSideOrderStateError) as err:
                    self.logger.error(f"Skipping order cancel: {err}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


def is_emitting_trading_signals(ctx):
    return ctx.is_trading_signal_emitter()


async def emit_trading_signals(ctx):
    try:
        return await ctx.emit_signal()
    except Exception as e:
        ctx.logger.exception(e, True, f"Error when emitting trading signal: {e}")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.asset as asset_class


class SpotAsset(asset_class.Asset):
    def __eq__(self, other):
        if isinstance(other, SpotAsset):
            return self.available == other.available and self.total == other.total
        return False

    def update(self, available=constants.ZERO, total=constants.ZERO):
        """
        Update asset portfolio
        :param available: the available delta
        :param total: the total delta
        :return: True if updated
        """
        if available == constants.ZERO and total == constants.ZERO:
            return False

        with self.update_or_restore():
            self.available += self._ensure_update_validity(self.available, available)
            self.total += self._ensure_update_validity(self.total, total)
            return True

    def set(self, available, total):
        """
        Set available and total values for portfolio asset
        :param available: the available value
        :param total: the total value
        :return: True if updated
        """
        if available == self.available and total == self.total:
            return False
        self.available = available
        self.total = total
        return True

    def reset(self):
        """
        Reset asset portfolio to zero
        """
        self.set(available=constants.ZERO, total=constants.ZERO)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import typing

import trading_backend

import octobot_commons.authentication as authentication
import octobot_trading.errors as errors
import octobot_trading.exchanges as exchanges


async def create_exchanges(exchange_manager, exchange_config_by_exchange: typing.Optional[dict[str, dict]]):
    if exchange_manager.is_sandboxed and not exchange_manager.exchange_only:
        exchange_manager.logger.info(f"Using sandbox exchange for {exchange_manager.exchange_name}")

    if exchange_manager.is_backtesting:
        # simulated : create exchange simulator instance
        await create_simulated_exchange(exchange_manager)
        exchange_manager.load_constants()
    else:
        # real : create a rest or websocket exchange instance
        await create_real_exchange(exchange_manager, exchange_config_by_exchange)
        exchange_manager.load_constants()
        await initialize_real_exchange(exchange_manager)

    if not exchange_manager.exchange_only:
        # create exchange producers if necessary
        await exchanges.create_exchange_producers(exchange_manager)

    if exchange_manager.is_backtesting:
        await init_simulated_exchange(exchange_manager)

    exchange_manager.exchange_name = exchange_manager.exchange.name
    exchange_manager.is_ready = True


async def create_real_exchange(exchange_manager, exchange_config_by_exchange: typing.Optional[dict[str, dict]]) -> None:
    """
    Create and initialize real REST exchange
    :param exchange_manager: the related exchange manager
    :param exchange_config_by_exchange: optional exchange configurations
    """
    await _create_rest_exchange(exchange_manager, exchange_config_by_exchange)
    try:
        await exchange_manager.exchange.initialize()
        _create_exchange_backend(exchange_manager)
        if exchange_manager.exchange_only:
            return
        await _initialize_exchange_backend(exchange_manager)
    except errors.AuthenticationError as err:
        if (
            exchange_manager.without_auth or exchange_manager.exchange.REQUIRES_AUTHENTICATION
            or exchange_manager.disable_unauth_retry
        ):
            # auth is required or already retried, don't loop
            exchange_manager.logger.warning(
                f"Authentication is required and raised an error: impossible to connect to "
                f"{exchange_manager.exchange_name} exchange: {err}"
            )
            raise
        exchange_manager.logger.error(
            f"Authentication error on {exchange_manager.exchange_name}, retrying without authentication..."
        )
        exchange_manager.without_auth = True
        await create_real_exchange(exchange_manager, exchange_config_by_exchange)
        return


async def initialize_real_exchange(exchange_manager):
    if not exchange_manager.exchange_only:
        await exchanges.create_exchange_channels(exchange_manager)

    # create Websocket exchange if possible
    if not exchange_manager.rest_only:
        # search for websocket
        if exchanges.check_web_socket_config(exchange_manager.config, exchange_manager.exchange.name):
            await exchanges.search_and_create_websocket(exchange_manager)


def _create_exchange_backend(exchange_manager):
    try:
        exchange_manager.exchange_backend = trading_backend.exchange_factory.create_exchange_backend(
            exchange_manager.exchange
        )
    except Exception as e:
        exchange_manager.logger.exception(e, True, f"Error when creating exchange backend: {e}")


async def _initialize_exchange_backend(exchange_manager):
    if exchange_manager.exchange_backend is not None and exchange_manager.exchange.authenticated() \
            and not exchange_manager.is_trader_simulated:
        exchange_manager.logger.debug(await exchange_manager.exchange_backend.initialize())
        initial_is_broker_enabled = exchange_manager.is_broker_enabled
        try:
            exchange_manager.is_broker_enabled, _ = await exchange_manager.exchange_backend.is_valid_account()
            exchange_manager.logger.debug(f"Broker rebate enabled: {exchange_manager.is_broker_enabled}")
        except Exception as err:
            exchange_manager.is_broker_enabled = initial_is_broker_enabled
            exchange_manager.logger.debug(f"Error when checking account broker state: {err}")


async def _is_supporting_octobot() -> bool:
    try:
        authenticator = authentication.Authenticator.instance()
        if not authenticator.is_initialized():
            initialization_timeout = 5
            await authenticator.await_initialization(initialization_timeout)
        if authenticator.user_account.supports.is_supporting():
            return True
    except asyncio.TimeoutError:
        pass
    return False


async def _create_rest_exchange(
    exchange_manager, exchange_config_by_exchange: typing.Optional[dict[str, dict]]
) -> None:
    """
    create REST based on ccxt exchange
    :param exchange_manager: the related exchange manager
    """
    await _search_and_create_rest_exchange(exchange_manager, exchange_config_by_exchange)

    if not exchange_manager.exchange:
        raise Exception(f"Can't create an exchange instance that match the exchange configuration ({exchange_manager})")


async def create_simulated_exchange(exchange_manager):
    exchange_manager.exchange = exchanges.ExchangeSimulator(
        exchange_manager.config, exchange_manager, exchange_manager.backtesting
    )

    await exchange_manager.exchange.initialize()
    _initialize_simulator_time_frames(exchange_manager)
    exchange_manager.exchange_config.set_config_time_frame()
    exchange_manager.exchange_config.set_config_traded_pairs()
    await exchanges.create_exchange_channels(exchange_manager)


async def init_simulated_exchange(exchange_manager):
    await exchange_manager.exchange.create_backtesting_exchange_producers()


async def _search_and_create_rest_exchange(
    exchange_manager, exchange_config_by_exchange: typing.Optional[dict[str, dict]]
) -> None:
    """
    Create a rest exchange if a RestExchange matching class is found
    :param exchange_manager: the related exchange manager
    """
    rest_exchange_class = exchanges.get_rest_exchange_class(
        exchange_manager.exchange_name, exchange_manager.tentacles_setup_config, exchange_config_by_exchange
    )
    if rest_exchange_class:
        if rest_exchange_class.HAS_FETCHED_DETAILS:
            await rest_exchange_class.fetch_exchange_config(exchange_config_by_exchange, exchange_manager)
        exchange_manager.exchange = rest_exchange_class(
            config=exchange_manager.config, exchange_manager=exchange_manager,
            exchange_config_by_exchange=exchange_config_by_exchange
        )


def _initialize_simulator_time_frames(exchange_manager):
    """
    Initialize simulator client time frames
    :param exchange_manager: the related exchange manager
    """
    exchange_manager.client_time_frames = exchange_manager.exchange.get_available_time_frames()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.personal_data as personal_data


@pytest.fixture
def trader():
    return mock.Mock(exchange_manager=mock.Mock())


def test_create_trailing_profile():
    with pytest.raises(NotImplementedError):
        personal_data.create_trailing_profile("Plop", {})
    with pytest.raises(TypeError):
        personal_data.create_trailing_profile(
            personal_data.TrailingProfileTypes.FILLED_TAKE_PROFIT, {}
        )
    assert isinstance(personal_data.create_trailing_profile(
        personal_data.TrailingProfileTypes.FILLED_TAKE_PROFIT, {"steps": []}
    ), personal_data.FilledTakeProfitTrailingProfile)

    trailing_profile = personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(price, price, True)
        for price in (10000, 12000, 13000)
    ])
    assert personal_data.create_trailing_profile(
        personal_data.TrailingProfileTypes.FILLED_TAKE_PROFIT, trailing_profile.to_dict()
    ) == trailing_profile


def test_create_filled_take_profit_trailing_profile(trader):
    with pytest.raises(ValueError):
        personal_data.create_filled_take_profit_trailing_profile(decimal.Decimal("0.5"), [])
    order_1 = personal_data.Order(trader)
    order_1.trigger_above = True
    order_1.origin_price = decimal.Decimal(1)
    order_2 = personal_data.Order(trader)
    order_2.trigger_above = True
    order_2.origin_price = decimal.Decimal(2)
    trailing_profile = personal_data.create_filled_take_profit_trailing_profile(
        decimal.Decimal("0.5"), [order_1, order_2]
    )
    assert trailing_profile == personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(0.5, 1, True),
        personal_data.TrailingPriceStep(1, 2, True),
    ])

    order_1.trigger_above = False
    order_1.origin_price = decimal.Decimal(1)
    order_2 = personal_data.Order(trader)
    order_2.trigger_above = False
    order_2.origin_price = decimal.Decimal(2)
    trailing_profile = personal_data.create_filled_take_profit_trailing_profile(
        decimal.Decimal("0.5"), [order_1, order_2]
    )
    assert trailing_profile == personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(0.5, 1, False),
        personal_data.TrailingPriceStep(1, 2, False),
    ])

    # incompatible trigger_above values (can't be both True and False)
    order_2.trigger_above = True
    with pytest.raises(ValueError):
        personal_data.create_filled_take_profit_trailing_profile(decimal.Decimal("0.5"), [order_1, order_2])



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import decimal
import typing

import octobot_trading.personal_data.portfolios.asset as asset
import octobot_trading.personal_data.portfolios.assets
import octobot_trading.personal_data.orders.order as order_import
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.portfolios.sub_portfolio_data as sub_portfolio_data
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.errors as errors

import octobot_commons.constants as commons_constants
import octobot_commons.logging as commons_logging
import octobot_commons.symbols as symbol_util

import numpy as numpy


def parse_decimal_portfolio(portfolio, as_decimal=True):
    decimal_portfolio = {}
    for symbol, symbol_balance in portfolio.items():
        if isinstance(symbol_balance, dict):
            decimal_portfolio[symbol] = {}
            portfolio_to_fill = decimal_portfolio[symbol]
            for balance_type, balance_val in symbol_balance.items():
                if isinstance(balance_val, (int, float, decimal.Decimal)):
                    portfolio_to_fill[balance_type] = decimal.Decimal(str(balance_val)) \
                        if as_decimal else float(balance_val)
    return decimal_portfolio


def parse_decimal_config_portfolio(portfolio):
    return {
        symbol: {
            k: decimal.Decimal(str(v))
            for k, v in symbol_balance.items()
        } if isinstance(symbol_balance, dict) else decimal.Decimal(str(symbol_balance))
        for symbol, symbol_balance in portfolio.items()
    }


def filter_empty_values(portfolio):
    return {
        symbol: value
        for symbol, value in portfolio.items()
        if value[commons_constants.PORTFOLIO_TOTAL] > 0
    }


def portfolio_to_float(portfolio, use_wallet_balance_on_futures=False):
    float_portfolio = {}
    for symbol, symbol_balance in portfolio.items():
        if (
            isinstance(symbol_balance, octobot_trading.personal_data.portfolios.assets.FutureAsset)
            and use_wallet_balance_on_futures
        ):
            float_portfolio[symbol] = {
                commons_constants.PORTFOLIO_AVAILABLE: float(symbol_balance.available),
                commons_constants.PORTFOLIO_TOTAL: float(symbol_balance.wallet_balance)
            }
        elif isinstance(symbol_balance, asset.Asset):
            float_portfolio[symbol] = {
                commons_constants.PORTFOLIO_AVAILABLE: float(symbol_balance.available),
                commons_constants.PORTFOLIO_TOTAL: float(symbol_balance.total)
            }
    return float_portfolio


def format_dict_portfolio_values(
    portfolio: dict[str, dict[str, typing.Union[float, decimal.Decimal]]], as_decimal: bool
) -> dict[str, dict[str, typing.Union[float, decimal.Decimal]]]:
    return {
        asset_name: {
            key: decimal.Decimal(str(val)) if as_decimal else float(val)
            for key, val in asset_content.items()
        }
        for asset_name, asset_content in portfolio.items()
    }


def get_draw_down(exchange_manager):
    """
    Draw down is the lowest portfolio value in % ever reached during a run
    :return: the Draw down value
    """
    draw_down = constants.ZERO
    if exchange_manager.is_future:
        try:
            value_currency = exchange_manager.exchange_personal_data.portfolio_manager.reference_market
            draw_down_pair = exchange_manager.exchange_config.traded_symbol_pairs[0]
            if exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
                draw_down_pair,
                enums.PositionSide.BOTH
            ).symbol_contract.is_inverse_contract():
                value_currency = symbol_util.parse_symbol(draw_down_pair).base
            if exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio \
                    is None:
                return constants.ZERO
            origin_value = exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder \
                .origin_portfolio.portfolio[value_currency].total
            portfolio_history = [origin_value]
            for transaction in exchange_manager.exchange_personal_data.transactions_manager.transactions.values():
                current_pnl = transaction.quantity if hasattr(transaction, "quantity") else transaction.realised_pnl
                portfolio_history.append(portfolio_history[-1] + current_pnl)

                current_draw_down = constants.ONE_HUNDRED - \
                    (portfolio_history[-1] / (max(portfolio_history) / constants.ONE_HUNDRED))

                draw_down = current_draw_down if current_draw_down > draw_down else draw_down
        except Exception as e:
            commons_logging.get_logger(__name__).warning(f"Error when computing draw down: {e}")
    return draw_down


async def get_coefficient_of_determination_data(transactions, start_balance,
                                                use_high_instead_of_end_balance=True,
                                                x_as_trade_count=True):
    if transactions:
        # get realized portfolio history

        pnl_history = [start_balance]
        pnl_history_times = []
        for transaction in transactions:
            current_pnl = None
            if hasattr(transaction, "quantity"):
                current_pnl = float(transaction.quantity)
                pnl_history_times.append(transaction.creation_time)
            elif hasattr(transaction, 'realised_pnl'):
                current_pnl = float(transaction.realised_pnl)
                pnl_history_times.append(transaction.creation_time)
            elif isinstance(transaction, dict):
                if transaction["quantity"]:
                    current_pnl = transaction["quantity"]
                    pnl_history_times.append(transaction["x"])
                elif transaction['realised_pnl']:
                    current_pnl = transaction['realised_pnl']
                    pnl_history_times.append(transaction["x"])

            if current_pnl is not None:
                pnl_history.append(pnl_history[-1] + current_pnl)

        if pnl_history_times:

            pnl_history_for_every_candle = None
            # either use trade to trade basis or candle to candle basis
            if x_as_trade_count is True:
                start_time = pnl_history_times[0]
                pnl_history_times.insert(0, start_time)
                end_time = pnl_history_times[-1]
                data_length = len(pnl_history)

            else:  # calculate pnl history for every candle
                raise NotImplementedError("x_as_trade_count=False is not implemented")
            # calculate best case data (exponential growth)
            end_balance = pnl_history[-1]

            if start_balance > end_balance:
                # if we end up with a negative balance we can't compute this
                return None, None, None, None, None

            if use_high_instead_of_end_balance:
                end_value = max(pnl_history)
            else:  # use the end balance
                end_value = end_balance

            def get_ideal_value(linear_growth, adj1, adj2):
                return ((linear_growth + adj1) ** pw) * adj2

            x = [start_time, end_time]
            y = [start_balance, end_value]

            pw = 15
            A = numpy.exp(numpy.log(y[0] / y[1]) / pw)
            a = (x[0] - x[1] * A) / (A - 1)
            b = y[0] / (x[0] + a) ** pw

            linear_growth = numpy.linspace(start_time, end_time, data_length)
            best_case = get_ideal_value(linear_growth, a, b)
            pnl_data = pnl_history_for_every_candle or pnl_history
            return list(best_case), pnl_data, start_balance, end_balance, pnl_history_times
    return None, None, None, None, None


async def get_coefficient_of_determination(exchange_manager, use_high_instead_of_end_balance=True):
    """
    Calculates proximity to the best case growth for the current run (best growth being associated to an exponential
    curve). The closer the actual result, the higher the coefficient_of_determination (R squared) will be.
    Return 0 if we end up with less money that we had to begin with
    :param use_high_instead_of_end_balance: best case exponential growth based on end balance or highest balance
    """
    coefficient_of_determination = 0
    if exchange_manager.exchange_personal_data.portfolio_manager is None or \
            exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio is None:
        return 0
    # get data the data necessary to compute the coefficient_of_determination
    origin_portfolio = portfolio_to_float(
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio.portfolio
    )
    start_balance = origin_portfolio[
        exchange_manager.exchange_personal_data.portfolio_manager.reference_market][commons_constants.PORTFOLIO_TOTAL]

    best_case, pnl_data, start_balance, _, _ \
        = await get_coefficient_of_determination_data(transactions=exchange_manager.exchange_personal_data.
                                                      transactions_manager.transactions.values(),
                                                      start_balance=start_balance,
                                                      use_high_instead_of_end_balance=use_high_instead_of_end_balance)

    if pnl_data:
        # calculate r²
        corr_matrix = numpy.corrcoef(best_case, pnl_data)
        corr = corr_matrix[0, 1]
        coefficient_of_determination = corr ** 2

    return round(coefficient_of_determination, 3)


def get_asset_price_from_converter_or_tickers(
    exchange_manager, to_convert_asset: str, target_asset: str, symbol: str, tickers: dict
):
    # 1. try with converter
    try:
        price = exchange_manager.exchange_personal_data.portfolio_manager. \
            portfolio_value_holder.value_converter.evaluate_value(
                to_convert_asset, constants.ONE, raise_error=True,
                target_currency=target_asset, init_price_fetchers=False
            )
        if price == constants.ZERO:
            raise errors.MissingPriceDataError
    except errors.MissingPriceDataError:
        # 2. try with tickers
        try:
            price = decimal.Decimal(str(
                tickers[symbol][enums.ExchangeConstantsTickersColumns.CLOSE.value]
                or tickers[symbol][enums.ExchangeConstantsTickersColumns.PREVIOUS_CLOSE.value]
            ))
        except KeyError:
            price = None
    return price


def resolve_sub_portfolios(
    master_portfolio: sub_portfolio_data.SubPortfolioData,
    sub_portfolios: list[sub_portfolio_data.SubPortfolioData],
    allowed_filling_assets: list[str],
    forbidden_filling_assets: list[str],
    market_prices: dict[str, float],
) -> (sub_portfolio_data.SubPortfolioData, list[sub_portfolio_data.SubPortfolioData]):
    resolved_portfolios = []
    remaining_master_portfolio = copy.deepcopy(master_portfolio)
    for sub_portfolio in sorted(sub_portfolios, key=lambda x: x.priority_key):
        resolved_portfolio = _resolve_sub_portfolio(
            remaining_master_portfolio, sub_portfolio, allowed_filling_assets, forbidden_filling_assets, market_prices
        )
        resolved_portfolios.append(resolved_portfolio)
    return remaining_master_portfolio, resolved_portfolios


def _resolve_sub_portfolio(
    remaining_master_portfolio: sub_portfolio_data.SubPortfolioData,
    sub_portfolio: sub_portfolio_data.SubPortfolioData,
    allowed_filling_assets: list[str],
    forbidden_filling_assets: list[str],
    market_prices: dict[str, float],
) -> sub_portfolio_data.SubPortfolioData:
    resolved_content = {}
    funds_deltas = {}
    missing_funds = {}
    missing_amount_by_asset = {}
    forecasted_sub_portfolio_content = sub_portfolio.get_content_after_deltas()
    for asset_name, holdings in forecasted_sub_portfolio_content.items():
        missing_amount_in_asset = _resolve_sub_portfolio_asset(
            asset_name, holdings, remaining_master_portfolio, resolved_content,
            sub_portfolio.locked_funds_by_asset.get(asset_name, constants.ZERO), forbidden_filling_assets
        )
        if missing_amount_in_asset > constants.ZERO:
            missing_amount_by_asset[asset_name] = missing_amount_in_asset
    # now that assets have been allocated, try to compensate missing assets
    for asset_name, missing_amount_in_asset in missing_amount_by_asset.items():
        _fill_missing_assets_from_allowed_filling_assets(
            asset_name, allowed_filling_assets, market_prices,
            remaining_master_portfolio, missing_amount_in_asset,
            resolved_content, funds_deltas, missing_funds
        )

    return sub_portfolio_data.SubPortfolioData(
        sub_portfolio.bot_id,
        sub_portfolio.portfolio_id,
        sub_portfolio.priority_key,
        resolved_content,
        sub_portfolio.unit,
        funds_deltas,
        missing_funds,
    )


def _resolve_sub_portfolio_asset(
    asset_name: str,
    holdings: dict[str, decimal.Decimal],
    remaining_master_portfolio: sub_portfolio_data.SubPortfolioData,
    resolved_content: dict[str, dict[str, decimal.Decimal]],
    locked_funds: decimal.Decimal,
    forbidden_filling_assets: list[str]
) -> decimal.Decimal:
    required_total_amount = holdings[commons_constants.PORTFOLIO_TOTAL]
    required_available_amount = holdings[commons_constants.PORTFOLIO_TOTAL] - locked_funds
    missing_amount_in_asset = constants.ZERO
    if asset_name in remaining_master_portfolio.content and asset_name not in forbidden_filling_assets:
        remaining_master_holding = remaining_master_portfolio.content[asset_name]
        # available = total - locked_funds
        remaining_total = remaining_master_holding[commons_constants.PORTFOLIO_TOTAL]
        remaining_available = remaining_master_holding[commons_constants.PORTFOLIO_AVAILABLE]
        allowed_missing_ratio_multiplier = constants.ONE - constants.SUB_PORTFOLIO_ALLOWED_MISSING_RATIO
        if remaining_total < locked_funds * allowed_missing_ratio_multiplier:
            # should never happen, log it in case it does
            commons_logging.get_logger(__name__).error(
                f"Unexpected missing locked {asset_name} funds in total portfolio {remaining_total=} {locked_funds=}"
            )
        # As orders are still open, their funds must be locked. Therefore, only consider available funds
        if remaining_available >= required_available_amount * allowed_missing_ratio_multiplier:
            if remaining_total < required_total_amount * allowed_missing_ratio_multiplier:
                # should never happen, log it in case it does
                commons_logging.get_logger(__name__).error(
                    f"Unexpected missing total {asset_name} value in portfolio: ensure order funds really locked"
                )
            sub_portfolio_total_amount = min(remaining_total, required_total_amount)
            # Do not use remaining_available or remaining_available here to avoid side effects due to locked fees
            # on some exchanges and not on others. Only compute from total holdings.
            # ex: coinbase locks 101 USDC to buy for 100 USDC + 1 USDC fee,
            # binance will just lock 100 and given less of the bought asset
            sub_portfolio_available_amount = sub_portfolio_total_amount - locked_funds
            # assets are available in master portfolio: reduce it from master
            if asset_name in resolved_content:
                resolved_content[asset_name][commons_constants.PORTFOLIO_TOTAL] += sub_portfolio_total_amount
                resolved_content[asset_name][commons_constants.PORTFOLIO_AVAILABLE] += sub_portfolio_available_amount
            else:
                resolved_content[asset_name] = {
                    commons_constants.PORTFOLIO_TOTAL: sub_portfolio_total_amount,
                    commons_constants.PORTFOLIO_AVAILABLE: sub_portfolio_available_amount,
                }
            remaining_master_portfolio.content[asset_name][commons_constants.PORTFOLIO_TOTAL] -= sub_portfolio_total_amount
            # don't make remaining available negative
            removed_portfolio_available_amount = min(remaining_available, sub_portfolio_available_amount)
            remaining_master_portfolio.content[asset_name][commons_constants.PORTFOLIO_AVAILABLE] -= removed_portfolio_available_amount
        else:
            # As orders are still open, their funds must be locked. Therefore, only consider available funds
            # Not enough assets in master portfolio
            usable_total = min(remaining_total, remaining_available + locked_funds)
            if asset_name in resolved_content:
                resolved_content[asset_name][commons_constants.PORTFOLIO_TOTAL] += usable_total
                resolved_content[asset_name][commons_constants.PORTFOLIO_AVAILABLE] += remaining_available
            else:
                resolved_content[asset_name] = {
                    commons_constants.PORTFOLIO_TOTAL: usable_total,
                    commons_constants.PORTFOLIO_AVAILABLE: remaining_available,
                }
            missing_amount_in_asset = abs(remaining_available - required_available_amount)
            remaining_master_portfolio.content[asset_name][commons_constants.PORTFOLIO_TOTAL] -= usable_total
            remaining_master_portfolio.content[asset_name][commons_constants.PORTFOLIO_AVAILABLE] -= remaining_available
    else:
        if locked_funds > constants.ZERO:
            # should never happen, log it in case it does
            commons_logging.get_logger(__name__).error(
                f"Unexpected missing {asset_name} value in portfolio but {locked_funds=}"
            )
        resolved_content[asset_name] = {
            commons_constants.PORTFOLIO_TOTAL: constants.ZERO,
            commons_constants.PORTFOLIO_AVAILABLE: constants.ZERO,
        }
        # asset not in master portfolio (anymore):
        missing_amount_in_asset = required_available_amount
    return missing_amount_in_asset


def _fill_missing_assets_from_allowed_filling_assets(
    asset_name, allowed_filling_assets, market_prices,
    remaining_master_portfolio, missing_amount_in_asset, resolved_content, funds_deltas, missing_funds
):
    # missing funds: try to get it from allowed assets
    for filling_asset, price in _iterate_filling_assets(
        asset_name, allowed_filling_assets, market_prices
    ):
        if filling_asset in remaining_master_portfolio.content and price != constants.ZERO:
            # use available funds only here to avoid taking funds locked in orders
            available_funds_in_filling_asset = remaining_master_portfolio.content[filling_asset][
                commons_constants.PORTFOLIO_AVAILABLE
            ]
            ideal_funds_to_add_in_filling_asset = missing_amount_in_asset * price
            if available_funds_in_filling_asset >= ideal_funds_to_add_in_filling_asset:
                # use a part of available filling asset to compensate missing amount in sub portfolio
                funds_to_add_in_filling_asset = ideal_funds_to_add_in_filling_asset
                taken_filling_asset = ideal_funds_to_add_in_filling_asset
                missing_amount_in_asset = constants.ZERO
            else:
                # use all available filling asset to compensate missing amount in sub portfolio
                funds_to_add_in_filling_asset = available_funds_in_filling_asset
                taken_filling_asset = available_funds_in_filling_asset
                missing_amount_in_asset = missing_amount_in_asset - (funds_to_add_in_filling_asset / price)
            if filling_asset in resolved_content:
                resolved_content[filling_asset][commons_constants.PORTFOLIO_TOTAL] += (
                    funds_to_add_in_filling_asset
                )
                resolved_content[filling_asset][commons_constants.PORTFOLIO_AVAILABLE] += (
                    funds_to_add_in_filling_asset
                )
            else:
                resolved_content[filling_asset] = {
                    commons_constants.PORTFOLIO_TOTAL: funds_to_add_in_filling_asset,
                    commons_constants.PORTFOLIO_AVAILABLE: funds_to_add_in_filling_asset,
                }
            if filling_asset in funds_deltas:
                funds_deltas[filling_asset][commons_constants.PORTFOLIO_TOTAL] += funds_to_add_in_filling_asset
                funds_deltas[filling_asset][commons_constants.PORTFOLIO_AVAILABLE] += funds_to_add_in_filling_asset
            elif funds_to_add_in_filling_asset != constants.ZERO:
                # only add delta when delta is not 0
                funds_deltas[filling_asset] = {
                    commons_constants.PORTFOLIO_TOTAL: funds_to_add_in_filling_asset,
                    commons_constants.PORTFOLIO_AVAILABLE: funds_to_add_in_filling_asset
                }
            remaining_master_portfolio.content[filling_asset][commons_constants.PORTFOLIO_AVAILABLE] -= (
                taken_filling_asset
            )
            remaining_master_portfolio.content[filling_asset][commons_constants.PORTFOLIO_TOTAL] -= (
                taken_filling_asset
            )
        if missing_amount_in_asset == constants.ZERO:
            return constants.ZERO
    if missing_amount_in_asset > constants.ZERO:
        # can't compensate for missing funds
        if asset_name in missing_funds:
            missing_funds[asset_name] += missing_amount_in_asset
        else:
            missing_funds[asset_name] = missing_amount_in_asset
    return missing_amount_in_asset


def _iterate_filling_assets(
    asset_name: str,
    allowed_filling_assets: list[str],
    tickers: dict[str, float],
):
    for allowed_filling_asset in allowed_filling_assets:
        symbol = symbol_util.merge_currencies(asset_name, allowed_filling_asset)
        if symbol in tickers:
            try:
                yield allowed_filling_asset, decimal.Decimal(str(tickers[symbol]))
            except decimal.DecimalException as err:
                commons_logging.get_logger(__name__).warning(
                    f"Error when reading {symbol} price: \"{tickers[symbol]}\": {err}"
                )
        else:
            reversed_symbol = symbol_util.merge_currencies(allowed_filling_asset, asset_name)
            if reversed_symbol in tickers and tickers[reversed_symbol]:
                yield allowed_filling_asset, constants.ONE / decimal.Decimal(str(tickers[reversed_symbol]))

def get_portfolio_filled_orders_deltas(
    previous_portfolio_content: dict[str, dict[str, decimal.Decimal]],
    updated_portfolio_content: dict[str, dict[str, decimal.Decimal]],
    filled_orders: list[order_import.Order]
) -> (dict[str, dict[str, decimal.Decimal]], dict[str, dict[str, decimal.Decimal]]):
    orders_asset_deltas, expected_fee_related_deltas, possible_fees_asset_deltas = get_assets_delta_from_orders(filled_orders)
    portfolios_asset_deltas = _get_assets_delta_from_portfolio(previous_portfolio_content, updated_portfolio_content)
    # return portfolio asset deltas that can approximately be explained by given orders
    orders_linked_deltas = {}
    ignored_deltas = {}
    for asset_name, holdings_delta in portfolios_asset_deltas.items():
        if asset_name not in orders_asset_deltas:
            # this delta is not due to these orders: skip it
            continue
        order_asset_delta = orders_asset_deltas[asset_name]
        holding_total = holdings_delta[commons_constants.PORTFOLIO_TOTAL]
        min_allowed_equivalent_total_holding = (
            holding_total * (constants.ONE - constants.SUB_PORTFOLIO_ALLOWED_DELTA_RATIO)
        )
        max_allowed_equivalent_total_holding = (
            holding_total * (constants.ONE + constants.SUB_PORTFOLIO_ALLOWED_DELTA_RATIO)
        )
        if (
            order_asset_delta * min_allowed_equivalent_total_holding < 0
            and (order_asset_delta + possible_fees_asset_deltas.get(asset_name, constants.ZERO))
                * min_allowed_equivalent_total_holding < 0
        ):
            # different sign: incompatible, this is unexpected
            ignored_deltas[asset_name] = holdings_delta
        elif (
            # approx same value
            abs(min_allowed_equivalent_total_holding)
            < abs(order_asset_delta)
            < abs(max_allowed_equivalent_total_holding)
        ):
            # similar delta (+/- fees)
            # this delta is due to these orders: add it
            orders_linked_deltas[asset_name] = holdings_delta
        elif asset_name in expected_fee_related_deltas and (
            abs(expected_fee_related_deltas[asset_name]) > abs(holdings_delta[commons_constants.PORTFOLIO_AVAILABLE])
        ):
            # paid fees are higher than holding delta: mostly paid fees, validate this delta
            orders_linked_deltas[asset_name] = holdings_delta
        elif (
            # try while taking expected and lower than expected fees into account
            asset_name in expected_fee_related_deltas and (
                any(
                    abs(min_allowed_equivalent_total_holding)
                    # fees bring back the order delta into the expected portfolio delta window
                    < abs(order_asset_delta + expected_fee)
                    < abs(max_allowed_equivalent_total_holding)
                    for expected_fee in (
                        # try with smaller fees in case user is paying less fees (when having large volume)
                        # test from 10% up to 100% fees
                        expected_fee_related_deltas[asset_name] * decimal.Decimal(multiplier / 10)
                        for multiplier in range(1, 11)
                    )
                )
            )
        ):
            # similar delta considering fees
            # this delta is due to these orders: add it
            orders_linked_deltas[asset_name] = holdings_delta
        elif abs(order_asset_delta) > abs(max_allowed_equivalent_total_holding):
            # last chance: try with potential fees, in case the expected fee currency was wrong
            if (
                # try while taking possible fees into account
                asset_name in possible_fees_asset_deltas and (
                    abs(min_allowed_equivalent_total_holding)
                    # fees bring back the order delta into the expected portfolio delta window
                    < abs(order_asset_delta + possible_fees_asset_deltas[asset_name])
                    < abs(max_allowed_equivalent_total_holding)
                )
            ):
                orders_linked_deltas[asset_name] = holdings_delta
            else:
                # too little in portfolio delta to be from those orders
                # As potential fees have already been taken into account, THIS IS UNEXPECTED.
                # Add it to ignored_deltas
                ignored_deltas[asset_name] = holdings_delta
        elif abs(min_allowed_equivalent_total_holding) > abs(order_asset_delta):
            # too much in portfolio delta: only take what is linked to the orders deltas
            # => updates both total and available
            # Should very rarely happen as it might reduce the total portfolio if done when unecessary
            commons_logging.get_logger(__name__).warning(
                f"Too large portfolio {asset_name} delta: {holdings_delta}, reducing to order delta: {order_asset_delta}"
            )
            orders_linked_deltas[asset_name] = {
                key: order_asset_delta
                for key in holdings_delta
            }
    for asset_name, order_delta in orders_asset_deltas.items():
        can_have_no_delta = (
            (order_delta == constants.ZERO)
            or (order_delta + possible_fees_asset_deltas.get(asset_name, constants.ZERO)) == constants.ZERO
            or (order_delta - possible_fees_asset_deltas.get(asset_name, constants.ZERO)) == constants.ZERO
        )
        if asset_name not in portfolios_asset_deltas and not can_have_no_delta:
            # asset not in portfolio delta, this is unexpected, register it in ignored deltas
            ignored_deltas[asset_name] = {
                commons_constants.PORTFOLIO_TOTAL: order_delta,
                commons_constants.PORTFOLIO_AVAILABLE: order_delta
            }
    return orders_linked_deltas, ignored_deltas


def _get_assets_delta_from_portfolio(
    previous_portfolio_content: dict,
    updated_portfolio_content: dict,
) ->  dict[str, dict[str, decimal.Decimal]]:
    asset_deltas = {}
    for asset_name in set(previous_portfolio_content).union(updated_portfolio_content):
        if previous_portfolio_content.get(asset_name) != updated_portfolio_content.get(asset_name):
            if asset_name not in previous_portfolio_content:
                # asset has been added
                asset_deltas[asset_name] = updated_portfolio_content[asset_name]
            elif asset_name not in updated_portfolio_content:
                # asset has been removed
                asset_deltas[asset_name] = {key: -val for key, val in previous_portfolio_content[asset_name].items()}
            else:
                # asset was already here
                asset_deltas[asset_name] = {}
                for key, updated_value in updated_portfolio_content[asset_name].items():
                    asset_deltas[asset_name][key] = updated_value - previous_portfolio_content[asset_name][key]
    return asset_deltas


def get_assets_delta_from_orders(orders: list[order_import.Order]) -> (
    dict[str, decimal.Decimal], dict[str, decimal.Decimal], dict[str, decimal.Decimal]
):
    asset_deltas = {}
    expected_fee_related_deltas = {}
    possible_fee_related_deltas = {}
    for order in orders:
        base, quote = symbol_util.parse_symbol(order.symbol).base_and_quote()
        # order "expected" related deltas
        added_unit_and_amount = (
            (base, order.origin_quantity)
            if order.side is enums.TradeOrderSide.BUY else (quote, order.total_cost)
        )
        removed_unit_and_amount = (
            (quote, order.total_cost)
            if order.side is enums.TradeOrderSide.BUY else (base, order.origin_quantity)
        )
        for unit_and_amount, multiplier in zip(
            (added_unit_and_amount, removed_unit_and_amount),
            (1, -1)
        ):
            if unit_and_amount[0] not in asset_deltas:
                asset_deltas[unit_and_amount[0]] = unit_and_amount[1] * multiplier
            else:
                asset_deltas[unit_and_amount[0]] += unit_and_amount[1] * multiplier
        # order "probable" related deltas (account for worse case fees)
        expected_forecasted_fees = order.get_computed_fee(use_origin_quantity_and_price=True)
        possible_forecasted_fees = _get_other_asset_forecasted_fees(order, expected_forecasted_fees)
        for fee_asset in (base, quote):
            for fee in (expected_forecasted_fees, possible_forecasted_fees):
                is_expected_fee = fee is expected_forecasted_fees
                if asset_amount := order_util.get_fees_for_currency(fee, fee_asset):
                    for delta_counter in (expected_fee_related_deltas, possible_fee_related_deltas):
                        if delta_counter is expected_fee_related_deltas and not is_expected_fee:
                            # when updating expected_fee_related_deltas, only account for expected fees
                            continue
                        if fee_asset not in delta_counter:
                            delta_counter[fee_asset] = -asset_amount
                        else:
                            delta_counter[fee_asset] += -asset_amount
    return asset_deltas, expected_fee_related_deltas, possible_fee_related_deltas

def _get_other_asset_forecasted_fees(order: order_import.Order, forecasted_fees: dict) -> dict:
    base, quote = symbol_util.parse_symbol(order.symbol).base_and_quote()
    other_fee = copy.deepcopy(forecasted_fees)
    if base_fee := order_util.get_fees_for_currency(forecasted_fees, base):
        other_fee[enums.FeePropertyColumns.CURRENCY.value] = quote
        other_fee[enums.FeePropertyColumns.COST.value] = base_fee * order.origin_price
    elif quote_fee := order_util.get_fees_for_currency(forecasted_fees, quote):
        other_fee[enums.FeePropertyColumns.CURRENCY.value] = base
        other_fee[enums.FeePropertyColumns.COST.value] = quote_fee / order.origin_price
    return other_fee



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.logging as logging

import octobot_trading.constants
import octobot_trading.enums as enums
import octobot_trading.errors
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.personal_data.state as state_class


class OrderState(state_class.State):
    MAX_SYNCHRONIZATION_ATTEMPTS = -1   # implement self._force_final_state() when this is enabled

    def __init__(self, order, is_from_exchange_data, enable_associated_orders_creation=True):
        super().__init__(is_from_exchange_data)

        # ensure order has not been cleared
        self.ensure_not_cleared(order)

        # related order
        self.order = order

        self.enable_associated_orders_creation = enable_associated_orders_creation

    def is_created(self) -> bool:
        """
        :return: True if the Order is created
        """
        return True

    def is_open(self) -> bool:
        """
        :return: True if the Order is considered as open
        """
        return not (self.is_filled() or self.is_canceled() or self.is_closed())

    def is_filled(self) -> bool:
        """
        :return: True if the Order is considered as filled
        """
        return False

    def is_canceled(self) -> bool:
        """
        :return: True if the Order is considered as canceled
        """
        return False

    def is_max_synchronization_attempts_reached(self) -> bool:
        """
        :return: True if the MAX_SYNCHRONIZATION_ATTEMPTS is set and has been reached
        """
        return -1 < self.__class__.MAX_SYNCHRONIZATION_ATTEMPTS <= self.synchronization_attempts

    def get_logger(self):
        """
        :return: the order logger
        """
        return logging.get_logger(self.order.get_logger_name() if self.order is not None else
                                  f"{self.__class__.__name__}_without_order")

    def log_event_message(self, state_message, error=None):
        """
        Log an order state event
        """
        if state_message is enums.StatesMessages.ALREADY_SYNCHRONIZING:
            self.get_logger().debug(f"Trying to update a refreshing state for order: {self.order}")
        elif state_message is enums.StatesMessages.SYNCHRONIZING_ERROR:
            self.get_logger().exception(error, True, f"Error when synchronizing order {self.order}: {error}")
        else:
            exchange_name = self.order.exchange_manager.exchange_name if self.order.exchange_manager \
                else 'unknown exchange'
            self.get_logger().info(f"{self.order} {state_message.value} on {exchange_name}")

    async def should_be_updated(self) -> bool:
        """
        Defines if the instance should be updated
        :return: True when the order type is supported by the exchange
        """
        return self.order.is_self_managed()

    def allows_new_status(self, status) -> bool:
        """
        :return: True if the given status is compatible with the current state
        """
        return True

    async def replace_order(self, new_order):
        async with self.refresh_operation():
            self.order.state = None
            self.order = new_order
            self.order.state = self

    def _force_final_state(self):
        raise NotImplementedError("_force_final_state is not implemented")

    async def _synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Ask OrdersChannel Internal producer to refresh the order from the exchange
        :param force_synchronization: When True, for the update of the order from the exchange
        :return: the result of OrdersProducer.update_order_from_exchange()
        """
        if self.is_max_synchronization_attempts_reached():
            self.get_logger().info(
                f"Forcing {self.__class__.__name__} final state after {self.MAX_SYNCHRONIZATION_ATTEMPTS} "
                f"synchronization attempts for order {self.order}"
            )
            self._force_final_state()
            asyncio.create_task(self.on_refresh_successful())
        else:
            try:
                self.ensure_not_cleared(self.order)
                self.get_logger().info(
                    f"Synchronizing [{self._underlying_refreshed_state.value}] order {self.order} "
                    f"with {self.order.exchange_manager.exchange_name} exchange"
                )
                self.synchronization_attempts += 1
                await exchange_channel.get_chan(
                    octobot_trading.constants.ORDERS_CHANNEL,
                    self.order.exchange_manager.id
                ).get_internal_producer().update_order_from_exchange(
                    order=self.order,
                    wait_for_refresh=True,
                    force_job_execution=force_synchronization,
                )
            except octobot_trading.errors.InvalidOrderState:
                self.get_logger().debug(f"Skipping exchange synchronisation as order has already been closed.")

    async def wait_for_next_state(self, timeout) -> None:
        # terminate is called at the end of the state for most order states
        await self.wait_for_terminate(timeout)

    @staticmethod
    def ensure_not_cleared(order):
        if order.is_cleared():
            raise octobot_trading.errors.InvalidOrderState(f"Order has already been cleared. Order: {order}")

    def _is_synchronization_enabled(self):
        return self.order.is_synchronization_enabled()

    def clear(self):
        """
        Clear references
        """
        self.order = None
